import pandas as pd
from sklearn.model_selection import StratifiedKFold, KFold

from automl_tools.functions.imputation import get_dataset_missing_processing, get_dataset_binding_processing
from automl_tools.functions.logger import seed_everything, get_stage, get_logger
from automl_tools.functions.modeling import run_model
from automl_tools.functions.optimization import get_dataset_parameter_baseline
from automl_tools.functions.prepare import get_dataset_remove_low_information, get_dataset_remove_correlation
from automl_tools.functions.processing import get_dataset_info, get_dataset_resume_table, get_dataset_statistical, get_dataset_outlier, get_dataset_outlier_processing
from automl_tools.functions.selection import get_dataset_feature_selection
from automl_tools.functions.utils import reduce_memory, display_importances, display_roc_curve, display_precision_recall, ks_report

logger = get_logger()


def automl_run(train=pd.DataFrame,
               test=pd.DataFrame,
               id_col=None,
               target_col="target",
               imp_num="knn",
               imp_cat="knn",
               processing=None,
               mutual_information=False,
               correlation_drop=False,
               model_feature_selection=None,
               model_run="LR",
               augmentation=False,
               Stratified=True):
    """
    :type train: DataFrame
    :type test: DataFrame
    :type id_col: String -> Default: None
    :type target_col: String
    :type imp_num: String ["gaussian", "arbitrary", "median", "mean", "random", "knn"] -> Default: "knn"
    :type imp_cat: String ["frequent", "constant", "rare", "knn"] -> Default: "knn"
    :type processing: String : ["woe", "binding" ]-> Default: False
    :type mutual_information: Boolean : [True, False]-> Default: False
    :type correlation_drop: Boolean : [True, False]-> Default: False
    :type model_feature_selection: List :
        paramter:
            binary:
                  LR  : LogisticRegression
                  RF  : RandomForestClassifier
                  SVM : SVC
                  LS  : LASSO
                  RD  : RIDGE
                  NET : Elasticnet
                  DT  : DecisionTreeClassifier
                  ET  : ExtraTreesClassifier
                  GB  : GradientBoostingClassifier
                  AB  : AdaBoostClassifier
                  XGB  : XGBClassifier
                  LGB  : LGBMClassifier
                  CTB  : CatBoostClassifier
                  NGB  : NGBClassifier
        Default: ["LR", "RF", "LGB"]

    :type model_run: String :
        paramter:
            binary:
                  LR  : LogisticRegression
                  RF  : RandomForestClassifier
                  SVM : SVC
                  LS  : LASSO
                  RD  : RIDGE
                  NET : Elasticnet
                  DT  : DecisionTreeClassifier
                  ET  : ExtraTreesClassifier
                  GB  : GradientBoostingClassifier
                  AB  : AdaBoostClassifier
                  XGB  : XGBClassifier
                  LGB  : LGBMClassifier
                  CTB  : CatBoostClassifier
                  NGB  : NGBClassifier
        Default: ["RF"]
    :type augmentation: Boolean : [True, False]-> Default: False
    :type increase_positive: Boolean : [True, False]-> Default: False
    :type Stratified: Boolean : [True, False]-> Default: True
    """

    if model_feature_selection is None:
        model_feature_selection = ["LR", "RF", "LGB"]
    assert type(model_feature_selection) == list
    if 0 < len(model_feature_selection) <= 3:
        print(f"Model selection variables default: {model_feature_selection}")
    else:
        print(f"model_feature_selection required min 1 parameter and max 3")

    seed_everything()

    train = reduce_memory(train)
    test = reduce_memory(test)
    get_stage(0)
    get_stage(1)
    logger.info('Load processing')
    get_dataset_info(train, test, target_col)

    get_stage(2)
    _train, _test, _id_train, _id_test, _target = get_dataset_statistical(train, test, id_col, target_col)
    get_stage(3)
    get_dataset_resume_table(train, test)

    get_stage(4)
    get_stage(5)
    get_dataset_outlier(_train, _test)

    get_stage(6)
    _train, _test = get_dataset_outlier_processing(_train, _test)
    get_dataset_outlier(_train, _test)

    get_stage(7)
    logger.info('Load imputation')
    _train, _test = get_dataset_missing_processing(_train, _test, _target, imp_num=imp_num, imp_cat=imp_cat)
    get_dataset_resume_table(_train, _test)

    if processing:
        get_stage(9)
        _train, _test = get_dataset_binding_processing(_train, _test, _target, processing)
        get_dataset_outlier(_train, _test)

    if mutual_information:
        get_stage(10)
        _train, _test = get_dataset_remove_low_information(_train, _test)
    if correlation_drop:
        _train, _test = get_dataset_remove_correlation(_train, _test, _target, threshold=0.70)

    get_stage(11)
    get_stage(12)
    logger.info('Load feature selection')
    feature_name = get_dataset_feature_selection(_train, _target, model_names=model_feature_selection)

    get_stage(13)
    get_stage(14)
    logger.info('Load feature optimization')
    best_parameter_tuning = get_dataset_parameter_baseline(_train, _target, feature_name, model_run)

    get_stage(15)
    logger.info('Load cross-validation')
    if Stratified:
        folds = StratifiedKFold(n_splits=5, shuffle=True, random_state=2020)
    else:
        folds = KFold(n_splits=5, shuffle=True, random_state=2020)

    logger.info('Load run model')
    oof_preds, df_oof_train_preds, df_oof_test_preds, importances, score = run_model(
        _train, _test, _target, _id_train, _id_test, folds, augmentation,
        feature_name, model_run, best_parameter_tuning)

    logger.info('generate report')
    folds_idx = [(trn_idx, val_idx) for trn_idx, val_idx in folds.split(_train, _target)]
    display_importances(feature_importance_df_=importances, model_name=model_run)
    display_roc_curve(y_=_target, oof_preds_=oof_preds, folds_idx_=folds_idx, model_name=model_run)
    display_precision_recall(y_=_target, oof_preds_=oof_preds, folds_idx_=folds_idx, model_name=model_run)

    df_oof_train_preds.to_csv("trained.csv", index=False)
    df_oof_test_preds.to_csv("submission.csv", index=False)

    _ks_report = ks_report(data=df_oof_train_preds, target="target", prob="prediction")
    _ks_report.to_csv("ks_report.csv", index=False)



import gc

import numpy as np
import pandas as pd
from sklearn.metrics import roc_auc_score

from automl_tools.functions.selection import search_model
from automl_tools.functions.utils import target_count_class, find_optimal_cutoff


def augment(x, y, t=2):
    xs, xn = [], []
    for i in range(t):
        mask = y > 0
        x1 = x[mask].copy()
        ids = np.arange(x1.shape[0])
        for c in range(x1.shape[1]):
            np.random.shuffle(ids)
            x1[:, c] = x1[ids][:, c]
        xs.append(x1)

    for i in range(t // 2):
        mask = y == 0
        x1 = x[mask].copy()
        ids = np.arange(x1.shape[0])
        for c in range(x1.shape[1]):
            np.random.shuffle(ids)
            x1[:, c] = x1[ids][:, c]
        xn.append(x1)

    xs = np.vstack(xs)
    xn = np.vstack(xn)
    ys = np.ones(xs.shape[0])
    yn = np.zeros(xn.shape[0])
    x = np.vstack([x, xs, xn])
    y = np.concatenate([y, ys, yn])
    return x, y


def run_model(train_df, test_df, target, id_train, id_test, folds_,
              augmentation, feature_name, model_name, best_parameter_tuning):
    oof_preds = np.zeros(train_df.shape[0])
    sub_preds = np.zeros(test_df.shape[0])
    feature_importance_df = pd.DataFrame()

    feats = [col for col in train_df.columns if col in feature_name]
    model_type, _num_class = target_count_class(target)
    _clf = search_model(model_name, model_type, _num_class)

    for n_fold, (trn_idx, val_idx) in enumerate(folds_.split(train_df, target)):
        trn_x, trn_y = train_df[feats].iloc[trn_idx], target.iloc[trn_idx]
        val_x, val_y = train_df[feats].iloc[val_idx], target.iloc[val_idx]

        if augmentation:
            trn_x, trn_y = augment(trn_x.values, trn_y.values)
            trn_x = pd.DataFrame(trn_x)

        clf = _clf.set_params(**best_parameter_tuning)

        clf.fit(trn_x, trn_y)

        oof_preds[val_idx] = clf.predict_proba(val_x)[:, 1]
        sub_preds += clf.predict_proba(test_df[feats])[:, 1] / folds_.n_splits

        fold_importance_df = pd.DataFrame()
        fold_importance_df["feature"] = feats
        if model_name == "SVM":
            fold_importance_df["importance"] = clf.coef0
        elif model_name == "LR":
            fold_importance_df["importance"] = clf.coef_[0]
        elif model_name == "NGB":
            fold_importance_df["importance"] = clf.feature_importances_[0]
        else:
            fold_importance_df["importance"] = clf.feature_importances_
        fold_importance_df["fold"] = n_fold + 1
        feature_importance_df = pd.concat([feature_importance_df, fold_importance_df], axis=0)

        print(f'Fold {n_fold + 1:2d} auc : {roc_auc_score(val_y, oof_preds[val_idx]):.6f}')
        print(f'Fold {n_fold + 1:2d} cutoff : {find_optimal_cutoff(val_y, oof_preds[val_idx])}')

        del clf, trn_x, trn_y, val_x, val_y
        gc.collect()

    print('full auc score {:.6f}'.format(roc_auc_score(target, oof_preds)))
    print("full cutoff optimal: {0}".format(find_optimal_cutoff(target, oof_preds)))

    test_df['target'] = sub_preds

    df_oof_train_preds = pd.DataFrame({'id_code': id_train, 'target': target, 'prediction': oof_preds})
    df_oof_test_preds = pd.DataFrame({'id_code': id_test, 'prediction': sub_preds})
    return oof_preds, df_oof_train_preds, df_oof_test_preds, feature_importance_df, roc_auc_score(target, oof_preds)



import pandas as pd
import psutil
from catboost import CatBoostClassifier, CatBoostRegressor
from lightgbm import LGBMClassifier, LGBMRegressor
from ngboost import NGBClassifier, NGBRegressor
from ngboost.distns import Bernoulli
from ngboost.distns import Normal
from ngboost.distns import k_categorical
from prettytable import PrettyTable
from sklearn.ensemble import AdaBoostClassifier, AdaBoostRegressor
from sklearn.ensemble import ExtraTreesClassifier, ExtraTreesRegressor
from sklearn.ensemble import GradientBoostingClassifier, GradientBoostingRegressor
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
from sklearn.feature_selection import SelectFromModel
from sklearn.linear_model import SGDClassifier, SGDRegressor, LogisticRegression, LinearRegression
from sklearn.model_selection import train_test_split
from sklearn.svm import SVC, SVR
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor
from xgboost import XGBClassifier, XGBRegressor

from automl_tools.functions.utils import target_count_class, get_tabulate


def search_model(model_name, model_type, _num_class):
    global models
    n_jobs = psutil.cpu_count(logical=False) if psutil.cpu_count(logical=False) > 0 else None
    max_depth = 6
    num_leaves = 2 ^ max_depth - 1
    if model_type in ("binary", "multi_class"):
        XGB = XGBClassifier(objective="binary:logistic", eval_metric="error",
                            use_label_encoder=False, max_depth=max_depth,
                            n_estimators=100, learning_rate=0.05)
        LGB = LGBMClassifier(objective="binary", metric="logloss", num_class=_num_class,
                             max_depth=max_depth, num_leaves=num_leaves,
                             n_estimators=100, learning_rate=0.05)
        CTB = CatBoostClassifier(loss_function="Logloss", eval_metric="AUC", verbose=0,
                                 allow_writing_files=False,
                                 max_depth=max_depth, leaf_estimation_method='Newton',
                                 thread_count=n_jobs, learning_rate=0.05)
        NGB = NGBClassifier(Dist=Bernoulli, verbose=False)

        if model_name == "XGB":
            if model_type == "multi_class":
                _objective = "multi:softprob"
                _metric = 'merror'
                XGB.set_params(objective=_objective, eval_metric=_metric)
        elif model_name == "LGB":
            if model_type == "multi_class":
                _objective = "multiclass"
                _metric = 'multi_logloss'
                LGB.set_params(objective=_objective, metric=_metric)
        elif model_name == "CTB":
            if model_type == "multi_class":
                _objective = "MultiClass"
                _metric = 'AUC'
                CTB.set_params(objective=_objective, eval_metric=_metric)
        elif model_name == "NGB":
            if model_type == "multi_class":
                _objective = k_categorical(_num_class)
                NGB.set_params(Dist=_objective)

        models = dict(
            LR=LogisticRegression(),
            RF=RandomForestClassifier(max_depth=max_depth, n_estimators=100),
            SVM=SVC(probability=True),
            LS=SGDClassifier(loss="log", penalty="l1"),
            RD=SGDClassifier(loss="log", penalty="l2"),
            NET=SGDClassifier(loss="log", penalty="elasticnet"),
            DT=DecisionTreeClassifier(max_depth=max_depth),
            ET=ExtraTreesClassifier(max_depth=max_depth, n_estimators=100),
            GB=GradientBoostingClassifier(max_depth=max_depth, n_estimators=100),
            AB=AdaBoostClassifier(n_estimators=100),
            XGB=XGB,
            LGB=LGB,
            CTB=CTB,
            NGB=NGB
        )

    if model_type == "regression":
        XGB = XGBRegressor(objective="reg:squarederror", use_label_encoder=False, max_depth=max_depth,
                           num_leaves=num_leaves, n_estimators=100, learning_rate=0.05)
        LGB = LGBMRegressor(objective="regression", max_depth=max_depth, num_leaves=num_leaves,
                            n_estimators=100, learning_rate=0.05)
        CTB = CatBoostRegressor(loss_function="RMSE", max_depth=max_depth,
                                leaf_estimation_method='Newton', learning_rate=0.05)
        NGB = NGBRegressor(Dist=Normal, verbose=False)
        models = dict(
            LR=LinearRegression(),
            RF=RandomForestRegressor(max_depth=max_depth, n_estimators=100),
            SVM=SVR(kernel='rbf'),
            LS=SGDRegressor(loss="log", penalty="l1"),
            RD=SGDRegressor(loss="log", penalty="l2"),
            NET=SGDRegressor(loss="log", penalty="elasticnet"),
            DT=DecisionTreeRegressor(max_depth=max_depth),
            ET=ExtraTreesRegressor(max_depth=max_depth, n_estimators=100),
            GB=GradientBoostingRegressor(max_depth=max_depth, n_estimators=100),
            AB=AdaBoostRegressor(n_estimators=100),
            XGB=XGB,
            LGB=LGB,
            CTB=CTB,
            NGB=NGB
        )

    clf = models[model_name]
    return clf


def model_selection_support(train, target, model_names, model_type, _num_class):
    X_train, X_val, y_train, y_val = train_test_split(train, target, test_size=0.20, random_state=2021)

    feat_labels = train.columns.tolist()
    global model_based, clf
    selection_list = list()
    feature_name = list()

    for model_name in model_names:

        if model_type in ("binary", "multi_class"):
            clf = search_model(model_name, model_type, _num_class)
            model_based = SelectFromModel(clf, threshold='median')
        if model_type == "regression":
            clf = search_model(model_name, model_type, _num_class)
            model_based = SelectFromModel(clf, threshold='median')

        if model_name in "NGB":
            model_based.fit(X_train, y_train)
        else:
            model_based.fit_transform(X_train, y_train)

        feature_name += [feat_labels[col] for col in model_based.get_support(indices=True)]

        selection_dict = dict()
        for col in train.columns:
            if str(col) in feature_name:
                selection_dict[col] = "YES"
            else:
                selection_dict[col] = "NO"
        selection_dict["model_name"] = model_name
        selection_list.append(selection_dict)

    feature_name = list(set(feature_name))
    feature_count = len(feature_name)
    return selection_list, feature_name, feature_count


def get_dataset_feature_selection(train, target, model_names):
    model_type, _num_class = target_count_class(target)
    selection_list, feature_name, feature_count = model_selection_support(train, target, model_names, model_type, _num_class)
    df = pd.DataFrame(selection_list)
    df2 = df.set_index("model_name")
    # df2 = df2.reset_index()

    get_tabulate(df2)
    #
    # print(df2.head())
    # t = PrettyTable()
    # t.field_names = df2.columns.tolist()
    # for column, values in df2.iterrows():
    #     print(column, values)
    #     t.add_row(values.tolist())
    # print(t)
    print(f"Number feature selection: {feature_count}")

    return feature_name



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from IPython.core.display import display
from sklearn.metrics import roc_auc_score, roc_curve, average_precision_score, precision_recall_curve


def get_tabulate(t, transpose=False):
    t = pd.DataFrame(t)
    if transpose:
        display(t.T)
    else:
        display(t)
    return True


def zr_score_outlier(df):
    out = []
    med = np.median(df)
    for i in df:
        try:
            z = (0.6745 * (i - med)) / (np.median(ma))
        except:
            z = 0
        if np.abs(z) > 3:
            out.append(i)
    if len(out) > 0:
        has_out = "yes"
    else:
        has_out = "no"
    return has_out


def z_score_outlier(df):
    out = []
    m = np.mean(df)
    sd = np.std(df)
    for i in df:
        z = (i - m) / sd
        if np.abs(z) > 3:
            out.append(i)
    if len(out) > 0:
        has_out = "yes"
    else:
        has_out = "no"
    return has_out


def iqr_outliers(df):
    out = []
    q1 = df.quantile(0.25)
    q3 = df.quantile(0.75)
    iqr = q3 - q1
    Lower_tail = q1 - 1.5 * iqr
    Upper_tail = q3 + 1.5 * iqr
    for i in df:
        if i > Upper_tail or i < Lower_tail:
            out.append(i)
    if len(out) > 0:
        has_out = "yes"
    else:
        has_out = "no"
    return has_out


def winsorization_outliers(df):
    out = []
    q1 = np.percentile(df, 1)
    q3 = np.percentile(df, 99)
    for i in df:
        if i > q3 or i < q1:
            out.append(i)
    if len(out) > 0:
        has_out = "yes"
    else:
        has_out = "no"
    return has_out


def clean_inf_nan(df):
    return df.replace([np.inf, -np.inf], np.nan)


def reduce_memory(dataset, verbose=False):
    ds_tmp = dataset.copy()
    start_mem = ds_tmp.memory_usage().sum() / 1024 ** 2
    int_columns = ds_tmp.select_dtypes(include=[np.int8, np.int16, np.int32, np.int64]).columns.tolist()
    for col in int_columns:
        ds_tmp[col] = pd.to_numeric(arg=ds_tmp[col], downcast='integer')

    float_columns = ds_tmp.select_dtypes(include=[np.float32, np.float64]).columns.tolist()
    for col in float_columns:
        ds_tmp[col] = pd.to_numeric(arg=ds_tmp[col], downcast='float')

    end_mem = ds_tmp.memory_usage().sum() / 1024 ** 2
    ds_tmp = clean_inf_nan(ds_tmp)
    if verbose:
        print('Mem. usage decreased to {:5.2f} Mb ({:.1f} % reduction)'.format(end_mem, 100 * (start_mem - end_mem) / start_mem))
    del dataset
    return ds_tmp


def generate_id(dataset):
    dataset.insert(0, 'NEW_ID', range(1, 1 + len(dataset)))
    return dataset


def find_optimal_cutoff(target, predicted):
    fpr, tpr, threshold = roc_curve(target, predicted)
    i = np.arange(len(tpr))
    roc = pd.DataFrame({'tf': pd.Series(tpr - (1 - fpr), index=i), 'threshold': pd.Series(threshold, index=i)})
    roc_t = roc.iloc[(roc.tf - 0).abs().argsort()[:1]]
    return list(roc_t['threshold'])


def target_count_class(target):
    df = pd.DataFrame()
    df["target"] = target
    count_class = list(sorted(df["target"].value_counts().to_dict().keys()))
    count_class = len(count_class)
    if count_class == 2:
        count_class = 1
        model_type = "binary"
    elif 2 < count_class <= 10:
        model_type = "multi_class"
    else:
        model_type = "regression"
    return model_type, count_class


def display_roc_curve(y_, oof_preds_, folds_idx_, model_name):
    plt.figure(figsize=(6, 6))
    scores = []
    for n_fold, (_, val_idx) in enumerate(folds_idx_):
        fpr, tpr, thresholds = roc_curve(y_.iloc[val_idx], oof_preds_[val_idx])
        score = roc_auc_score(y_.iloc[val_idx], oof_preds_[val_idx])
        scores.append(score)
        plt.plot(fpr, tpr, lw=1, alpha=0.3, label='ROC fold %d (AUC = %0.4f)' % (n_fold + 1, score))

    plt.plot([0, 1], [0, 1], linestyle='--', lw=2, color='r', label='Luck', alpha=.8)
    fpr, tpr, thresholds = roc_curve(y_, oof_preds_)
    score = roc_auc_score(y_, oof_preds_)
    plt.plot(fpr, tpr, color='b', label=f'avg roc auc = {score:0.4f} $\pm$ {np.std(scores):0.4f})', lw=2, alpha=.8)

    plt.xlim([-0.05, 1.05])
    plt.ylim([-0.05, 1.05])
    plt.xlabel('False Positive Rate')
    plt.ylabel('True Positive Rate')
    plt.title(f'{model_name.upper()} ROC CURVE')
    plt.legend(loc="lower right")
    plt.tight_layout()
    plt.savefig(f'roc_curve_{model_name}.png')
    plt.show()


def display_precision_recall(y_, oof_preds_, folds_idx_, model_name):
    plt.figure(figsize=(6, 6))
    scores = []
    for n_fold, (_, val_idx) in enumerate(folds_idx_):
        fpr, tpr, thresholds = roc_curve(y_.iloc[val_idx], oof_preds_[val_idx])
        score = average_precision_score(y_.iloc[val_idx], oof_preds_[val_idx])
        scores.append(score)
        plt.plot(fpr, tpr, lw=1, alpha=0.3, label='AP fold %d (auc = %0.4f)' % (n_fold + 1, score))

    precision, recall, thresholds = precision_recall_curve(y_, oof_preds_)
    score = average_precision_score(y_, oof_preds_)
    plt.plot(precision, recall, color='b', lw=2, alpha=.8,
             label=f'avg roc (auc = {score:0.4f} $\pm$ {np.std(scores):0.4f})')

    plt.xlim([-0.05, 1.05])
    plt.ylim([-0.05, 1.05])
    plt.xlabel('Recall')
    plt.ylabel('Precision')
    plt.title(f'{model_name} Recall / Precision')
    plt.legend(loc="best")
    plt.tight_layout()
    plt.savefig(f'recall_precision_curve_{model_name}.png')
    plt.show()


def display_importances(feature_importance_df_, model_name):
    cols = feature_importance_df_[["feature", "importance"]].groupby("feature").mean() \
               .sort_values(by="importance", ascending=False)[:50].index
    best_features = feature_importance_df_.loc[feature_importance_df_.feature.isin(cols)]
    plt.figure(figsize=(8, 10))
    sns.barplot(x="importance", y="feature", data=best_features.sort_values(by="importance", ascending=False))
    plt.title(f'{model_name} Features (avg over folds)')
    plt.tight_layout()
    plt.savefig(f'{model_name}_importances.png')
    plt.show()


def ks_report(data=None, target=None, prob=None):
    data['target0'] = 1 - data[target]
    data['bucket'] = pd.qcut(data[prob], 10)
    grouped = data.groupby('bucket', as_index=False)
    kstable = pd.DataFrame()
    kstable['min_prob'] = grouped.min()[prob]
    kstable['max_prob'] = grouped.max()[prob]
    kstable['events'] = grouped.sum()[target]
    kstable['nonevents'] = grouped.sum()['target0']
    kstable = kstable.sort_values(by="min_prob", ascending=False).reset_index(drop=True)
    kstable['event_rate'] = (kstable.events / data[target].sum()).apply('{0:.2%}'.format)
    kstable['nonevent_rate'] = (kstable.nonevents / data['target0'].sum()).apply('{0:.2%}'.format)
    kstable['cum_eventrate'] = (kstable.events / data[target].sum()).cumsum()
    kstable['cum_noneventrate'] = (kstable.nonevents / data['target0'].sum()).cumsum()
    kstable['KS'] = np.round(kstable['cum_eventrate'] - kstable['cum_noneventrate'], 3) * 100

    # Formating
    kstable['cum_eventrate'] = kstable['cum_eventrate'].apply('{0:.2%}'.format)
    kstable['cum_noneventrate'] = kstable['cum_noneventrate'].apply('{0:.2%}'.format)
    kstable.index = range(1, 11)
    kstable.index.rename('Decile', inplace=True)

    # Display KS
    from colorama import Fore
    print(Fore.RED + "KS is " + str(max(kstable['KS'])) + "%" + " at decile " + str((kstable.index[kstable['KS'] == max(kstable['KS'])][0])))
    return kstable



import numpy as np
from prettytable import PrettyTable
from scipy import stats


def get_dataset_remove_low_information(train, test):
    # remove quasi constant
    t = PrettyTable()
    t.field_names = ['Quasi Variable']
    quasi_cols_remove = list()
    for col in train.columns:
        if train[col].std() == 0:
            quasi_cols_remove.append(col)

    t = PrettyTable()
    t.field_names = ['Column Quasi']
    if len(quasi_cols_remove) > 0:
        for col in quasi_cols_remove:
            t.add_row([col])
    else:
        t.add_row(["NOT VARIABLES QUASI"])
    print(t)

    # remove duplicate columns
    groups = train.columns.to_series().groupby(train.dtypes).groups
    duplicate_cols_remove = list()
    for t, v in groups.items():
        cs = train[v].columns
        vs = train[v]
        lcs = len(cs)
        for i in range(lcs):
            ia = vs.iloc[:, i].values
            for j in range(i + 1, lcs):
                ja = vs.iloc[:, j].values
                if np.array_equal(ia, ja):
                    duplicate_cols_remove.append(cs[i])
                    break

    t = PrettyTable()
    t.field_names = ['Column Duplicates']
    if len(duplicate_cols_remove) > 0:
        for col in duplicate_cols_remove:
            t.add_row([col])
    else:
        t.add_row(["NOT VARIABLES DUPLICATES"])
    print(t)

    # drop sparse columns
    flist = [x for x in train.columns]
    sparse_cols_remove = list()
    for col in flist:
        if len(np.unique(train[col])) < 2:
            sparse_cols_remove.append(col)

    t = PrettyTable()
    t.field_names = ['Column Sparse']
    if len(sparse_cols_remove) > 0:
        for col in sparse_cols_remove:
            t.add_row([col])
    else:
        t.add_row(["NOT VARIABLES SPARSE"])
    print(t)

    drop_all_columns = quasi_cols_remove + duplicate_cols_remove + sparse_cols_remove

    train.drop(drop_all_columns, axis=1, inplace=True)
    test.drop(drop_all_columns, axis=1, inplace=True)
    return train, test


def get_dataset_remove_correlation(train, test, target, threshold=0.95):
    drop_all_columns = list()

    for col in train.columns:
        value = stats.spearmanr(train[col].values, target.values)[0]
        if abs(value) > threshold:
            drop_all_columns.append(col)

    t = PrettyTable()
    t.field_names = ['Column Correlations']
    if len(drop_all_columns) > 0:
        for col in drop_all_columns:
            t.add_row([col])
    else:
        t.add_row(["NOT VARIABLES CORRELATIONS"])
    print(t)

    train.drop(drop_all_columns, axis=1, inplace=True)
    test.drop(drop_all_columns, axis=1, inplace=True)
    return train, test



import numpy as np
from hyperopt import STATUS_OK, hp, Trials, fmin, tpe, space_eval
from hyperopt.pyll import scope
from sklearn.model_selection import cross_val_score

from automl_tools.functions.selection import search_model
from automl_tools.functions.utils import target_count_class


def search_space(model):
    space = dict()
    if model == 'LR':
        space = dict(warm_start=hp.choice('warm_start', [True, False]),
                     fit_intercept=hp.choice('fit_intercept', [True, False]),
                     tol=hp.uniform('tol', 0.00001, 0.0001),
                     C=hp.uniform('C', 0.05, 3),
                     max_iter=hp.choice('max_iter', range(100, 1000))
                     )

    elif model == 'RF':
        space = dict(criterion='gini',
                     max_depth=6,
                     max_features=hp.choice('max_features', ['auto', 'sqrt', 'log2', None]),
                     min_samples_leaf=hp.uniform('min_samples_leaf', 0, 0.5),
                     min_samples_split=hp.uniform('min_samples_split', 0, 1),
                     n_estimators=100,
                     )

    elif model == 'SVM':
        space = dict(C=hp.uniform('C', 0.1, 2.0),
                     kernel=hp.choice('kernel', ['linear', 'poly', 'rbf', 'sigmoid']),
                     degree=scope.int(hp.quniform('degree', 2, 5, 1)),
                     gamma=hp.choice('gamma', ['auto', 'scale']),
                     probability=hp.choice('probability', [True]),
                     tol=hp.loguniform('tol', np.log(1e-5), np.log(1e-2))
                     )

    elif model == 'LS':
        space = dict(alpha=hp.uniform('alpha', 1, 10),
                     max_iter=hp.choice('max_iter', range(100, 1000)),
                     loss=hp.choice('loss', ['log', 'modified_huber', 'squared_loss',
                                             'huber', 'epsilon_insensitive']),
                     penalty=hp.choice('penalty', ['l1']),
                     learning_rate=hp.choice('learning_rate', ['optimal'])
                     )
    elif model == 'RD':
        space = dict(alpha=hp.uniform('alpha', 1, 10),
                     max_iter=hp.choice('max_iter', range(100, 1000)),
                     loss=hp.choice('loss', ['log', 'modified_huber', 'squared_loss',
                                             'huber', 'epsilon_insensitive']),
                     penalty=hp.choice('penalty', ['l2']),
                     learning_rate=hp.choice('learning_rate', ['optimal'])
                     )
    elif model == 'NET':
        space = dict(alpha=hp.uniform('alpha', 1, 10),
                     max_iter=hp.choice('max_iter', range(100, 1000)),
                     loss=hp.choice('loss', ['log', 'modified_huber', 'squared_loss',
                                             'huber', 'epsilon_insensitive']),
                     penalty=hp.choice('penalty', ['elasticnet']),
                     learning_rate=hp.choice('learning_rate', ['optimal'])
                     )

    elif model == "DT":
        space = dict(criterion=hp.choice('criterion', ['gini', 'entropy']),
                     splitter=hp.choice('splitter', ['best', 'random']),
                     max_depth=scope.int(hp.quniform('max_depth', 3, 10, 1)),
                     min_samples_split=scope.int(hp.quniform('min_samples_split', 2, 50, 1)),
                     min_samples_leaf=scope.int(hp.quniform('min_samples_leaf', 1, 50, 1)),
                     max_features=hp.choice('max_features', ['auto', 'log2', None])
                     )

    elif model == 'ET':
        space = dict(criterion=hp.choice('criterion', ['entropy', 'gini']),
                     max_depth=hp.choice('max_depth', np.arange(3, 12, dtype=int)),
                     max_features=hp.choice('max_features', ['auto', 'sqrt', 'log2', None]),
                     min_samples_leaf=hp.uniform('min_samples_leaf', 0, 0.5),
                     min_samples_split=hp.uniform('min_samples_split', 0, 1),
                     n_estimators=hp.choice('n_estimators', [10, 50]))

    elif model == 'GB':
        space = dict(learning_rate=hp.lognormal('learning_rate', 0.05, 0.3),
                     max_depth=hp.quniform('max_depth', 10, 12, 10),
                     max_features=hp.choice('max_features', ['auto', 'sqrt', 'log2', None]),
                     min_samples_leaf=hp.uniform('min_samples_leaf', 0, 0.5),
                     min_samples_split=hp.uniform('min_samples_split', 0, 1),
                     n_estimators=hp.choice('n_estimators', [10, 50])
                     )
    elif model == 'AB':
        space = dict(n_estimators=hp.choice('n_estimators', range(50, 100)),
                     learning_rate=hp.quniform('learning_rate', 0, 0.05, 0.0001)
                     )

    elif model == 'XGB':
        space = dict(learning_rate=hp.quniform('learning_rate', 0, 0.05, 0.0001),
                     n_estimators=hp.choice('n_estimators', range(100, 1000)),
                     eta=hp.quniform('eta', 0.025, 0.5, 0.005),
                     max_depth=hp.choice('max_depth', np.arange(2, 12, dtype=int)),
                     min_child_weight=hp.quniform('min_child_weight', 1, 9, 0.025),
                     subsample=hp.quniform('subsample', 0.5, 1, 0.005),
                     gamma=hp.quniform('gamma', 0.5, 1, 0.005),
                     colsample_bytree=hp.quniform('colsample_bytree', 0.5, 1, 0.005)
                     )
    elif model == 'LGB':
        space = dict(learning_rate=hp.quniform('learning_rate', 0, 0.05, 0.0001),
                     n_estimators=hp.choice('n_estimators', range(100, 1000)),
                     max_depth=hp.choice('max_depth', np.arange(2, 12, dtype=int)),
                     num_leaves=hp.choice('num_leaves', 2 * np.arange(2, 2 ** 11, dtype=int)),
                     min_child_weight=hp.quniform('min_child_weight', 1, 9, 0.025),
                     colsample_bytree=hp.quniform('colsample_bytree', 0.5, 1, 0.005)
                     )

    elif model == 'CTB':
        space = dict(n_estimators=hp.choice('n_estimators', np.arange(50, 250, 25)),
                     max_depth=hp.choice('max_depth', np.arange(2, 12, dtype=int)),
                     learning_rate=hp.quniform('learning_rate', 0, 0.05, 0.0001),
                     reg_lambda=hp.uniform('reg_lambda', 1, 10)
                     )

    elif model == 'NGB':
        space = dict(n_estimators=hp.choice('n_estimators', np.arange(50, 250, 25)),
                     learning_rate=hp.quniform('learning_rate', 0, 0.05, 0.0001),
                     minibatch_frac=hp.choice('minibatch_frac', [1.0, 0.5]),
                     tol=hp.uniform('tol', 0.00001, 0.0001),
                     )
    return space


def get_acc_status(clf, X, y):
    acc = cross_val_score(clf, X, y, cv=5).mean()
    return {'loss': -acc, 'status': STATUS_OK}


def best_search_tuning(model_name, X, y):
    model_type, _num_class = target_count_class(y)

    trials = Trials()

    space = search_space(model_name)
    if model_type == "regression":
        if model_name == "LR":
            space = dict(fit_intercept=hp.choice('fit_intercept', [True, False]),
                         normalizebool=hp.choice('normalizebool', [True, False]),
                         positivebool=hp.choice('positivebool', [True, False])
                         )

    def obj_fnc(params):
        clf = search_model(model_name, model_type, _num_class)
        clf.set_params(**params)
        return get_acc_status(clf, X, y)

    best = fmin(
        fn=obj_fnc,
        space=space,
        algo=tpe.suggest,
        max_evals=50,
        trials=trials
    )
    best = space_eval(space, best)
    return best


def get_dataset_parameter_baseline(train, target, feature_name, model_name):
    X = train[feature_name]
    y = target
    best_parameter_tuning = best_search_tuning(model_name, X, y)
    return best_parameter_tuning



import numpy as np
import pandas as pd
from category_encoders.cat_boost import CatBoostEncoder
from optbinning import BinningProcess, MulticlassOptimalBinning
from pandas.api.types import is_numeric_dtype
from pandas.api.types import is_string_dtype
from sklearn.ensemble import RandomForestRegressor
from sklearn.impute import KNNImputer
from sklearn.metrics import mean_squared_error
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import OrdinalEncoder

from automl_tools.functions.utils import target_count_class


def optimize_impute_knn(df, variable, target):
    rmse = lambda y, yhat: np.sqrt(mean_squared_error(y, yhat))

    errors_list = list()
    for k in range(1, 20, 2):
        imputer = KNNImputer(n_neighbors=k)
        imputed = imputer.fit_transform(df[variable].values.reshape(-1, 1))
        df_imputed = pd.DataFrame(imputed, columns=[variable])

        X = df_imputed
        y = target
        X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=2021)
        model = RandomForestRegressor()
        model.fit(X_train, y_train)
        preds = model.predict(X_test)
        error = rmse(y_test, preds)
        errors_list.append({'K': k, 'RMSE': error})
    optimal_k = min(errors_list, key=lambda x: x['RMSE'])
    return optimal_k["K"], errors_list


def impute_na(df_train, df_test, target,
              _type_numerical="gaussian",
              _type_categorical="knn"):
    def encoder_data(df):
        var_ord_enc = OrdinalEncoder()
        _var = df[col]
        _var_not_null = _var[_var.notnull()]
        _reshaped_vals = _var_not_null.values.reshape(-1, 1)
        _encoded_vals = var_ord_enc.fit_transform(_reshaped_vals)
        df.loc[_var.notnull(), col] = np.squeeze(_encoded_vals)
        return df, var_ord_enc

    for col in df_train.columns:
        if df_train[col].isnull().sum() > 0:
            if is_numeric_dtype(df_train[col]):

                if _type_numerical == "gaussian":
                    train_new_value = df_train[col].mean() + 3 * df_train[col].std()
                    test_new_value = df_test[col].mean() + 3 * df_test[col].std()
                    df_train[f'{col}_imputed'] = df_train[col].fillna(train_new_value)
                    df_test[f'{col}_imputed'] = df_test[col].fillna(test_new_value)

                elif _type_numerical == "arbitrary":
                    train_new_value = -999
                    test_new_value = -999
                    df_train[f'{col}_imputed'] = df_train[col].fillna(train_new_value)
                    df_test[f'{col}_imputed'] = df_test[col].fillna(test_new_value)

                elif _type_numerical == "median":
                    train_new_value = df_train[col].median()
                    test_new_value = df_test[col].median()
                    df_train[f'{col}_imputed'] = df_train[col].fillna(train_new_value)
                    df_test[f'{col}_imputed'] = df_test[col].fillna(test_new_value)

                elif _type_numerical == "mean":
                    train_new_value = df_train[col].mean()
                    test_new_value = df_test[col].mean()
                    df_train[f'{col}_imputed'] = df_train[col].fillna(train_new_value)
                    df_test[f'{col}_imputed'] = df_test[col].fillna(test_new_value)

                elif _type_numerical == "random":
                    train_random_sample = df_train[col].dropna().sample(df_train[col].isnull().sum(), random_state=2021)
                    test_random_sample = df_test[col].dropna().sample(df_test[col].isnull().sum(), random_state=2021)
                    df_train[f'{col}_imputed'] = np.where(df_train[col].isnull(), train_random_sample, df_train[col])
                    df_test[f'{col}_imputed'] = np.where(df_test[col].isnull(), test_random_sample, df_test[col])

                elif _type_numerical == "knn":
                    optimal_k, _ = optimize_impute_knn(df_train, col, target)
                    train_new_value = df_train[col]
                    test_new_value = df_test[col]
                    imputer = KNNImputer(n_neighbors=optimal_k)
                    train_imputed = imputer.fit_transform(train_new_value.values.reshape(-1, 1))
                    test_imputed = imputer.fit_transform(test_new_value.values.reshape(-1, 1))

                    df_train[f'{col}_imputed'] = pd.DataFrame(train_imputed)
                    df_test[f'{col}_imputed'] = pd.DataFrame(test_imputed)


            elif is_string_dtype(df_train[col]):

                if _type_categorical == "frequent":
                    train_new_value = df_train[col].mode()
                    test_new_value = df_test[col].mode()
                    df_train[f'{col}_imputed'] = df_train[col].fillna(train_new_value)
                    df_test[f'{col}_imputed'] = df_test[col].fillna(test_new_value)

                elif _type_categorical == "constant":
                    train_new_value = "missing"
                    test_new_value = "missing"
                    df_train[f'{col}_imputed'] = df_train[col].fillna(train_new_value)
                    df_test[f'{col}_imputed'] = df_test[col].fillna(test_new_value)

                elif _type_categorical == "rare":
                    _df_train = pd.DataFrame()
                    _df_train[col] = df_train[col]
                    total_houses = len(_df_train)
                    temp_df = pd.Series(_df_train[col].value_counts() / total_houses)
                    grouping_dict = {
                        k: ('rare' if k not in temp_df[temp_df >= 0.05].index else k)
                        for k in temp_df.index
                    }
                    df_train[f'{col}_imputed'] = df_train[col].map(grouping_dict)
                    df_test[f'{col}_imputed'] = df_test[col].map(grouping_dict)

                elif _type_categorical == "knn":
                    _df_train, _df_test = pd.DataFrame(), pd.DataFrame()
                    _df_train[col] = df_train[col]
                    _df_test[col] = df_test[col]

                    _df_train, var_ord_enc = encoder_data(_df_train)
                    _df_test, _ = encoder_data(_df_test)

                    optimal_k, _ = optimize_impute_knn(_df_train, col, target)
                    train_new_value = _df_train[col].copy()
                    test_new_value = _df_test[col].copy()
                    imputer = KNNImputer(n_neighbors=optimal_k)
                    train_imputed = imputer.fit_transform(train_new_value.values.reshape(-1, 1))
                    # train_inverse_imputed = var_ord_enc.inverse_transform(train_imputed)
                    test_imputed = imputer.fit_transform(test_new_value.values.reshape(-1, 1))
                    # test_inverse_imputed = var_ord_enc.inverse_transform(test_imputed)

                    df_train[f'{col}_imputed'] = pd.DataFrame(train_imputed)
                    df_test[f'{col}_imputed'] = pd.DataFrame(test_imputed)

    return df_train, df_test


def get_dataset_missing_processing(train, test, target, imp_num, imp_cat):
    train, test = impute_na(df_train=train, df_test=test,
                            target=target,
                            _type_numerical=imp_num,
                            _type_categorical=imp_cat)
    drop_all_columns = [str(col).split("_")[0] for col in train.columns.tolist() if str(col).endswith("_imputed")]

    cat_cols = [col for col in train.select_dtypes(include=['object']).columns]

    CBE_encoder = CatBoostEncoder()
    for col in cat_cols:
        train[col] = CBE_encoder.fit_transform(train[col], target)
        test[col] = CBE_encoder.transform(test[col])

    train.drop(drop_all_columns, axis=1, inplace=True)
    test.drop(drop_all_columns, axis=1, inplace=True)
    return train, test


def get_dataset_binding_processing(train, test, target, processing):
    model_type, _ = target_count_class(target)
    if processing == "binding":
        metric = "indices"
    else:
        metric = "woe"

    if str(model_type) == "binary":
        num_cols = [col for col in train.select_dtypes(include=['number']).columns]
        cat_cols = [col for col in train.select_dtypes(include=['object']).columns]
        all_var = num_cols + cat_cols
        X = train[all_var]
        y = target
        binning_process = BinningProcess(variable_names=all_var,
                                         categorical_variables=cat_cols)
        binning_process.fit(X, y)
        var_list = list(binning_process.get_support(names=True))
        for col in var_list:
            optb = binning_process.get_binned_variable(col)
            train_transform_indices = optb.transform(train[col].values, metric=metric)
            test_transform_indices = optb.transform(test[col].values, metric=metric)
            train[f'{col}_bin'] = train_transform_indices
            test[f'{col}_bin'] = test_transform_indices
            del train[col]
            del test[col]

    elif str(model_type) == "multi_class":
        y = target
        for col in train.columns:
            optb = MulticlassOptimalBinning(name=col, solver="mip")
            optb.fit(train[col], y)
            train_transform_indices = optb.transform(train[col].values, metric=metric)
            test_transform_indices = optb.transform(test[col].values, metric=metric)
            train[f'{col}_bin'] = train_transform_indices
            test[f'{col}_bin'] = test_transform_indices
            del train[col]
            del test[col]

    return train, test



from collections import OrderedDict

import numpy as np
import pandas as pd
from scipy import stats
from sklearn.preprocessing import LabelEncoder

from automl_tools.functions.utils import target_count_class, iqr_outliers, zr_score_outlier, winsorization_outliers, get_tabulate, generate_id


def get_dataset_info(train, test, target):
    t = [{"Train Row": train.shape[0],
          'Train Column': train.shape[1],
          'Test Row': test.shape[0],
          'Test Column': test.shape[1],
          'Test/Train ratio': f"{test.shape[0] / float(train.shape[0]):.2f}"
          }]
    get_tabulate(t, True)

    model_type, count_class = target_count_class(train[target])

    if model_type in ("binary", "multi_class"):
        print(f"Model type: {model_type}")
        t = list()
        count = OrderedDict(train[target].value_counts().to_dict())
        for key, value in count.items():
            t_dict = {"target": key, "#_rows": value, "%_perc": f'{value / train.shape[0]:.2f}'}
            t.append(t_dict)
        get_tabulate(t)
    else:
        print(f"Model type: {model_type}")
    return True


def get_dataset_statistical(train, test, id_col, target_col):
    train[target_col] = LabelEncoder().fit_transform(train[target_col])
    total_rows = train.shape[0]

    id_columns = []
    uniques_columns = []
    date_columns = []
    missing_columns = []
    numeric_binary_columns = []
    numeric_continuous_columns = []
    numeric_discrete_columns = []
    category_binary_columns = []
    category_nominal_columns = []

    if id_col is None:
        train = generate_id(train)
        test = generate_id(test)
        id_col = "NEW_ID"
    else:
        if not int(train[id_col].value_counts().shape[0]) == int(total_rows):
            raise ValueError(f'the column {id_col} is not unique value')

    id_columns.append(id_col)

    for col in train.columns:
        if train[col].dtype == 'datetime64[ns]':
            date_columns.append(col)

    for col in train.select_dtypes('number').columns:
        if train[col].value_counts().shape[0] == 2:
            numeric_binary_columns.append(col)
        elif 2 < train[col].value_counts().shape[0] <= 50:
            numeric_continuous_columns.append(col)
        elif int(train[col].value_counts().shape[0]) == int(total_rows):
            uniques_columns.append(col)
        else:
            numeric_discrete_columns.append(col)

    for col in train.select_dtypes('category').columns:
        if train[col].value_counts().shape[0] == 2:
            category_binary_columns.append(col)
        else:
            category_nominal_columns.append(col)

    for col in list(train):
        if (len(train[col].value_counts()) == 1) | (train[col].isnull().sum() / len(train) >= 0.90):
            missing_columns.append(col)

    t = [{"ID features": len(id_columns),
          'date features': len(date_columns),
          'missing features': len(missing_columns),
          'numeric Binary features': len(numeric_binary_columns),
          'numeric continuous features': len(numeric_continuous_columns),
          'numeric discrete features': len(numeric_discrete_columns),
          'categorical Binary features': len(category_binary_columns),
          'categorical Nominal features': len(category_nominal_columns)
          }]
    get_tabulate(t, True)

    _id_train = train[id_columns[0]]
    _target = train[target_col]
    drop_all_columns = missing_columns + id_columns + date_columns + [target_col] + uniques_columns

    train.drop(drop_all_columns, axis=1, inplace=True)
    _id_test = test[id_columns[0]]
    test = test[train.columns]
    return train, test, _id_train, _id_test, _target


def get_dataset_resume_table(train, test):
    train_total_values = train.shape[0]
    test_total_values = test.shape[0]

    t = list()

    for col in train.columns:
        train_dtype = train[col].dtypes
        train_unique_list = list(train[col].unique())
        train_total_uniques = len(train[col].unique())
        train_total_missing = train[col].isnull().sum()
        train_perc_missing = np.round((100 * train[col].isnull().sum() / train_total_values), 2)
        train_entropy = round(stats.entropy(train[col].value_counts(normalize=True), base=2), 2)

        if {col}.issubset(test.columns):
            test_dtype = test[col].dtypes
            test_unique_list = list(test[col].unique())
            test_total_uniques = len(test[col].unique())
            test_total_missing = test[col].isnull().sum()
            test_perc_missing = np.round((100 * test[col].isnull().sum() / test_total_values), 2)
            test_entropy = round(stats.entropy(test[col].value_counts(normalize=True), base=2), 2)
            train_with_test_unique = len(train_unique_list) - len(set(train_unique_list).intersection(test_unique_list))
            test_with_train_unique = len(test_unique_list) - len(set(test_unique_list).intersection(train_unique_list))
        else:
            test_dtype = np.nan
            test_total_uniques = np.nan
            test_total_missing = np.nan
            test_perc_missing = np.nan
            test_entropy = np.nan
            train_with_test_unique = np.nan
            test_with_train_unique = np.nan

        t_dict = dict(
            Name=col,
            train_dtype=train_dtype, train_uniques=train_total_uniques, train_missing=train_total_missing,
            train_perc_missing=train_perc_missing, train_entropy=train_entropy,
            test_dtype=test_dtype, test_uniques=test_total_uniques, test_missing=test_total_missing,
            test_perc_missing=test_perc_missing, test_entropy=test_entropy,
            train_with_test_unique=train_with_test_unique, test_with_train_unique=test_with_train_unique)
        t.append(t_dict)

    get_tabulate(t)


def get_dataset_outlier(train, test):
    train_total_values = train.shape[0]
    test_total_values = test.shape[0]

    t = list()
    for col in train.select_dtypes('number').columns:
        train_dtype = train[col].dtypes
        train_zr_score_outlier = zr_score_outlier(train[col])
        train_winsorization_outliers = winsorization_outliers(train[col])

        if {col}.issubset(test.columns):
            test_dtype = test[col].dtypes
            test_zr_score_outlier = zr_score_outlier(test[col])
            test_winsorization_outliers = winsorization_outliers(test[col])
        else:
            test_dtype = np.nan
            test_zr_score_outlier = np.nan
            test_winsorization_outliers = np.nan

        t_dict = dict(
            Name=col,
            train_dtype=train_dtype, train_total_values=train_total_values,
            train_zrscore=train_zr_score_outlier, train_winsorizercscore=train_winsorization_outliers,
            test_dtype=test_dtype, test_total_values=test_total_values,
            test_zrscore=test_zr_score_outlier, test_winsorizercscore=test_winsorization_outliers)
        t.append(t_dict)
    get_tabulate(t)
    return True


def get_dataset_outlier_processing(train, test):
    pd.options.mode.chained_assignment = None

    for col in train.select_dtypes('number').columns:
        is_out = iqr_outliers(train[col])
        if is_out == "yes":
            first_quartile, third_quartile = np.percentile(train[col], [25, 75])
            first_percentile, ninetynine_percentile = np.percentile(train[col], [1, 99])
            iqr = third_quartile - first_quartile
            lower_bound = first_quartile - (1.5 * iqr)
            upper_bound = third_quartile + (1.5 * iqr)
            train[col].loc[train[col] > upper_bound] = ninetynine_percentile
            test[col].loc[test[col] > upper_bound] = ninetynine_percentile
            train[col].loc[train[col] < lower_bound] = first_percentile
            test[col].loc[test[col] < lower_bound] = first_percentile

    return train, test



import gc
import logging
import os
import random
import warnings
from datetime import datetime

import numpy as np
import pandas as pd
from prettytable import PrettyTable


def get_logger():
    FORMAT = "[%(levelname)s]%(asctime)s:%(name)s:%(message)s"
    logging.basicConfig(format=FORMAT)
    logger = logging.getLogger('main')
    logger.setLevel(logging.DEBUG)
    return logger


def timer(start_time=None):
    if not start_time:
        start_time = datetime.now()
        return start_time
    elif start_time:
        tmin, tsec = divmod((datetime.now() - start_time).total_seconds(), 60)
        print(" Time taken: %i minutes and %s seconds." % (tmin, round(tsec, 2)))


def seed_everything(seed=2021):
    gc.enable()
    random.seed(seed)
    os.environ['PYTHONHASHSEED'] = str(seed)
    np.random.seed(seed)

    warnings.filterwarnings('always')
    warnings.filterwarnings('ignore')
    warnings.filterwarnings(action='ignore', category=DeprecationWarning)
    warnings.simplefilter(action='ignore', category=FutureWarning)

    pd.set_option('precision', 5)
    pd.set_option('display.max_columns', None)
    pd.set_option('display.expand_frame_repr', False)
    pd.set_option('max_colwidth', -1)


def get_stage(value):
    bcolors = dict(HEADER='\033[95m',
                   OKBLUE='\033[94m',
                   OKGREEN='\033[90m',
                   WARNING='\033[93m',
                   FAIL='\033[91m',
                   ENDC='\033[0m',
                   BOLD='\033[1m',
                   UNDERLINE='\033[4m')

    if value == 0:
        t = PrettyTable()
        t.field_names = [f'{bcolors.get("OKGREEN")}::::::::::::DATASET ANALYSIS ::::::::::::{bcolors.get("ENDC")}']
        print(t)
    elif value == 1:
        t = PrettyTable()
        t.field_names = [f'{bcolors.get("OKGREEN")}1.1 ::::::::::::DATASET ROWS  ::::::::::::{bcolors.get("ENDC")}']
        print(t)
    elif value == 2:
        t = PrettyTable()
        t.field_names = [f'{bcolors.get("OKGREEN")}1.2 ::::::::::::DATASET VARIABLES ANALYSIS ::::::::::::{bcolors.get("ENDC")}']
        print(t)
    elif value == 3:
        t = PrettyTable()
        t.field_names = [f'{bcolors.get("OKGREEN")}1.3 ::::::::::::DATASET RESUME ANALYSIS ::::::::::::{bcolors.get("ENDC")}']
        print(t)
    elif value == 4:
        t = PrettyTable()
        t.field_names = [f'{bcolors.get("OKGREEN")} ::::::::::::DATASET CLEANING ::::::::::::{bcolors.get("ENDC")}']
        print(t)
    elif value == 5:
        t = PrettyTable()
        t.field_names = [f'{bcolors.get("OKGREEN")}2.1 ::::::::::::DATASET OUTLIER::::::::::::{bcolors.get("ENDC")}']
        print(t)
    elif value == 6:
        t = PrettyTable()
        t.field_names = [f'{bcolors.get("OKGREEN")}2.1 ::::::::::::DATASET OUTLIER NUMERIC PROCESSING::::::::::::{bcolors.get("ENDC")}']
        print(t)
    elif value == 7:
        t = PrettyTable()
        t.field_names = [f'{bcolors.get("OKGREEN")}2.2 ::::::::::::DATASET MISSING PROCESSING::::::::::::{bcolors.get("ENDC")}']
        print(t)
    elif value == 8:
        t = PrettyTable()
        t.field_names = [f'{bcolors.get("OKGREEN")}2.2 ::::::::::::DATASET WOE PROCESSING::::::::::::{bcolors.get("ENDC")}']
        print(t)
    elif value == 9:
        t = PrettyTable()
        t.field_names = [f'{bcolors.get("OKGREEN")}2.2 ::::::::::::DATASET BINDING PROCESSING::::::::::::{bcolors.get("ENDC")}']
        print(t)
    elif value == 10:
        t = PrettyTable()
        t.field_names = [f'{bcolors.get("OKGREEN")}2.2 ::::::::::::DATASET CORRELATION PROCESSING::::::::::::{bcolors.get("ENDC")}']
        print(t)
    elif value == 11:
        t = PrettyTable()
        t.field_names = [f'{bcolors.get("OKGREEN")} ::::::::::::DATASET FEATURE SELECTION::::::::::::{bcolors.get("ENDC")}']
        print(t)
    elif value == 12:
        t = PrettyTable()
        t.field_names = [f'{bcolors.get("OKGREEN")}3.1 ::::::::::::DATASET COMPARE SELECTION::::::::::::{bcolors.get("ENDC")}']
        print(t)

    elif value == 13:
        t = PrettyTable()
        t.field_names = [f'{bcolors.get("OKGREEN")} ::::::::::::DATASET MODEL BASELINE::::::::::::{bcolors.get("ENDC")}']
        print(t)
    elif value == 14:
        t = PrettyTable()
        t.field_names = [f'{bcolors.get("OKGREEN")} ::::::::::::DATASET PARAMETER MODEL::::::::::::{bcolors.get("ENDC")}']
        print(t)
    elif value == 15:
        t = PrettyTable()
        t.field_names = [f'{bcolors.get("OKGREEN")} ::::::::::::DATASET RUN MODEL::::::::::::{bcolors.get("ENDC")}']
        print(t)
